@@ -1,13 +1,13 @@
+"""Define FlowEnsemble object that holds an ensemble of normalizing flows."""
 from typing import Any, Callable, Sequence, Tuple
 
 import dill as pickle
-import jax.numpy as np
-import numpy as onp
+import jax.numpy as jnp
+import numpy as np
 import pandas as pd
 from jax import random
-from jax.example_libraries.optimizers import Optimizer
 
-from pzflow import Flow
+from pzflow import Flow, distributions
 from pzflow.bijectors import Bijector_Info, InitFunction
 
 
@@ -20,27 +20,33 @@ class FlowEnsemble:
         List of DataFrame columns that the flows expect/produce.
     conditional_columns : tuple
         List of DataFrame columns on which the flows are conditioned.
+    latent: distributions.LatentDist
+        The latent distribution of the normalizing flows.
+        Has it's own sample and log_prob methods.
+    data_error_model : Callable
+        The error model for the data variables. See the docstring of
+        __init__ for more details.
+    condition_error_model : Callable
+        The error model for the conditional variables. See the docstring
+        of __init__ for more details.
     info : Any
         Object containing any kind of info included with the ensemble.
         Often Reverse the data the flows are trained on.
-    latent
-        The latent distribution of the normalizing flows.
-        Has it's own sample and log_prob methods.
     """
 
     def __init__(
         self,
         data_columns: Sequence[str] = None,
         bijector: Tuple[InitFunction, Bijector_Info] = None,
+        latent: distributions.LatentDist = None,
         conditional_columns: Sequence[str] = None,
-        latent=None,
         data_error_model: Callable = None,
         condition_error_model: Callable = None,
         autoscale_conditions: bool = True,
         N: int = 1,
         info: Any = None,
         file: str = None,
-    ):
+    ) -> None:
         """Instantiate an ensemble of normalizing flows.
 
         Note that while all of the init parameters are technically optional,
@@ -49,61 +55,65 @@ class FlowEnsemble:
 
         Parameters
         ----------
-        data_columns : Sequence[str], optional
+        data_columns : Sequence[str]; optional
             Tuple, list, or other container of column names.
             These are the columns the flows expect/produce in DataFrames.
-        bijector : Bijector Call, optional
+        bijector : Bijector Call; optional
             A Bijector call that consists of the bijector InitFunction that
             initializes the bijector and the tuple of Bijector Info.
             Can be the output of any Bijector, e.g. Reverse(), Chain(...), etc.
-        conditional_columns : Sequence[str], optional
-            Names of columns on which to condition the normalizing flows.
-        latent : distribution, optional
-            The latent distribution for the normalizing flows. Can be any of
+            If not provided, the bijector can be set later using
+            flow.set_bijector, or by calling flow.train, in which case the
+            default bijector will be used. The default bijector is
+            ShiftBounds -> RollingSplineCoupling, where the range of shift
+            bounds is learned from the training data, and the dimensions of
+            RollingSplineCoupling is inferred. The default bijector assumes
+            that the latent has support [-5, 5] for every dimension.
+        latent : distributions.LatentDist; optional
+            The latent distribution for the normalizing flow. Can be any of
             the distributions from pzflow.distributions. If not provided,
-            a normal distribution is used with the number of dimensions
-            inferred.
-        data_error_model : Callable, optional
+            a uniform distribution is used with input_dim = len(data_columns),
+            and B=5.
+        conditional_columns : Sequence[str]; optional
+            Names of columns on which to condition the normalizing flows.
+        data_error_model : Callable; optional
             A callable that defines the error model for data variables.
-            data_error_model must take key, X, Xerr, nsamples as arguments where:
-                key is a jax rng key, e.g. jax.random.PRNGKey(0)
-                X is a 2 dimensional array of data variables, where the order
-                    of variables matches the order of the columns in data_columns
-                Xerr is the corresponding 2 dimensional array of errors
-                nsamples is the number of samples to draw from the error distribution
+            data_error_model must take key, X, Xerr, nsamples as arguments:
+                - key is a jax rng key, e.g. jax.random.PRNGKey(0)
+                - X is 2D array of data variables, where the order of variables
+                    matches the order of the columns in data_columns
+                - Xerr is the corresponding 2D array of errors
+                - nsamples is number of samples to draw from error distribution
             data_error_model must return an array of samples with the shape
             (X.shape[0], nsamples, X.shape[1]).
-            If data_error_model is not provided, a Gaussian error model is assumed.
-        condition_error_model : Callable, optional
+            If data_error_model is not provided, Gaussian error model assumed.
+        condition_error_model : Callable; optional
             A callable that defines the error model for conditional variables.
-            condition_error_model must take key, X, Xerr, nsamples as arguments where:
-                key is a jax rng key, e.g. jax.random.PRNGKey(0)
-                X is a 2 dimensional array of conditional variables, where the order
-                    of variables matches the order of the columns in conditional_columns
-                Xerr is the corresponding 2 dimensional array of errors
-                nsamples is the number of samples to draw from the error distribution
-            condition_error_model must return an array of samples with the shape
+            condition_error_model must take key, X, Xerr, nsamples, where:
+                - key is a jax rng key, e.g. jax.random.PRNGKey(0)
+                - X is 2D array of conditional variables, where the order of
+                    variables matches order of columns in conditional_columns
+                - Xerr is the corresponding 2D array of errors
+                - nsamples is number of samples to draw from error distribution
+            condition_error_model must return array of samples with shape
             (X.shape[0], nsamples, X.shape[1]).
-            If condition_error_model is not provided, a Gaussian error model is assumed.
-        autoscale_conditions : bool, defautl=True
-            Sets whether or not conditions are automatically standard scaled when
-            passed to a conditional flow. I recommend you leave this as True.
-        N : int, default=1
+            If condition_error_model is not provided, Gaussian error model
+            assumed.
+        autoscale_conditions : bool; default=True
+            Sets whether or not conditions are automatically standard scaled
+            when passed to a conditional flow. I recommend you leave as True.
+        N : int; default=1
             The number of flows in the ensemble.
-        info : Any, optional
+        info : Any; optional
             An object to attach to the info attribute.
-        file : str, optional
+        file : str; optional
             Path to file from which to load a pretrained flow ensemble.
             If a file is provided, all other parameters must be None.
         """
 
         # validate parameters
-        if data_columns is None and bijector is None and file is None:
-            raise ValueError("You must provide data_columns and bijector OR file.")
-        if data_columns is not None and bijector is None:
-            raise ValueError("Please also provide a bijector.")
-        if data_columns is None and bijector is not None:
-            raise ValueError("Please also provide data_columns.")
+        if data_columns is None and file is None:
+            raise ValueError("You must provide data_columns OR file.")
         if file is not None and any(
             (
                 data_columns is not None,
@@ -141,8 +151,15 @@ class FlowEnsemble:
             # load the metadata
             self.data_columns = save_dict["data_columns"]
             self.conditional_columns = save_dict["conditional_columns"]
+            self.data_error_model = save_dict["data_error_model"]
+            self.condition_error_model = save_dict["condition_error_model"]
             self.info = save_dict["info"]
 
+            self._latent_info = save_dict["latent_info"]
+            self.latent = getattr(distributions, self._latent_info[0])(
+                *self._latent_info[1]
+            )
+
         # otherwise create a new ensemble from the provided parameters
         else:
             # save the ensemble of flows
@@ -163,6 +180,9 @@ class FlowEnsemble:
             # save the metadata
             self.data_columns = data_columns
             self.conditional_columns = conditional_columns
+            self.latent = self._ensemble["Flow 0"].latent
+            self.data_error_model = data_error_model
+            self.condition_error_model = condition_error_model
             self.info = info
 
     def log_prob(
@@ -171,7 +191,7 @@ class FlowEnsemble:
         err_samples: int = None,
         seed: int = None,
         returnEnsemble: bool = False,
-    ) -> np.ndarray:
+    ) -> jnp.ndarray:
         """Calculates log probability density of inputs.
 
         Parameters
@@ -181,15 +201,15 @@ class FlowEnsemble:
             Every column in self.data_columns must be present.
             If self.conditional_columns is not None, those must be present
             as well. If other columns are present, they are ignored.
-        err_samples : int, default=None
+        err_samples : int; default=None
             Number of samples from the error distribution to average over for
             the log_prob calculation. If provided, Gaussian errors are assumed,
             and method will look for error columns in `inputs`. Error columns
             must end in `_err`. E.g. the error column for the variable `u` must
             be `u_err`. Zero error assumed for any missing error columns.
-        seed : int, default=None
+        seed : int; default=None
             Random seed for drawing the samples with Gaussian errors.
-        returnEnsemble : bool, default=False
+        returnEnsemble : bool; default=False
             If True, returns log_prob for each flow in the ensemble as an
             array of shape (inputs.shape[0], N flows in ensemble).
             If False, the prob is averaged over the flows in the ensemble,
@@ -198,12 +218,12 @@ class FlowEnsemble:
 
         Returns
         -------
-        np.ndarray
+        jnp.ndarray
             For shape, see returnEnsemble description above.
         """
 
         # calculate log_prob for each flow in the ensemble
-        ensemble = np.array(
+        ensemble = jnp.array(
             [
                 flow.log_prob(inputs, err_samples, seed)
                 for flow in self._ensemble.values()
@@ -211,7 +231,7 @@ class FlowEnsemble:
         )
 
         # re-arrange so that (axis 0, axis 1) = (inputs, flows in ensemble)
-        ensemble = np.rollaxis(ensemble, axis=1)
+        ensemble = jnp.rollaxis(ensemble, axis=1)
 
         if returnEnsemble:
             # return the ensemble of log_probs
@@ -219,13 +239,13 @@ class FlowEnsemble:
         else:
             # return mean over ensemble
             # note we return log(mean prob) instead of just mean log_prob
-            return np.log(np.exp(ensemble).mean(axis=1))
+            return jnp.log(jnp.exp(ensemble).mean(axis=1))
 
     def posterior(
         self,
         inputs: pd.DataFrame,
         column: str,
-        grid: np.ndarray,
+        grid: jnp.ndarray,
         marg_rules: dict = None,
         normalize: bool = True,
         err_samples: int = None,
@@ -233,7 +253,7 @@ class FlowEnsemble:
         batch_size: int = None,
         returnEnsemble: bool = False,
         nan_to_zero: bool = True,
-    ) -> np.ndarray:
+    ) -> jnp.ndarray:
         """Calculates posterior distributions for the provided column.
 
         Calculates the conditional posterior distribution, assuming the
@@ -250,9 +270,9 @@ class FlowEnsemble:
             is calculated. Must be one of the columns in self.data_columns.
             However, whether or not this column is one of the columns in
             `inputs` is irrelevant.
-        grid : np.ndarray
+        grid : jnp.ndarray
             Grid on which to calculate the posterior.
-        marg_rules : dict, optional
+        marg_rules : dict; optional
             Dictionary with rules for marginalizing over missing variables.
             The dictionary must contain the key "flag", which gives the flag
             that indicates a missing value. E.g. if missing values are given
@@ -261,39 +281,39 @@ class FlowEnsemble:
             variables that will need to be marginalized over, where name is
             the name of the variable, and callable is a callable that takes
             the row of variables nad returns a grid over which to marginalize
-            the variable. E.g. {"y": lambda row: np.linspace(0, row["x"], 10)}.
+            the variable. E.g. {"y": lambda row: jnp.linspace(0, row["x"], 10)}.
             Note: the callable for a given name must *always* return an array
             of the same length, regardless of the input row.
-        normalize : boolean, default=True
+        normalize : boolean; default=True
             Whether to normalize the posterior so that it integrates to 1.
-        err_samples : int, default=None
+        err_samples : int; default=None
             Number of samples from the error distribution to average over for
             the posterior calculation. If provided, Gaussian errors are assumed,
             and method will look for error columns in `inputs`. Error columns
             must end in `_err`. E.g. the error column for the variable `u` must
             be `u_err`. Zero error assumed for any missing error columns.
-        seed : int, default=None
+        seed : int; default=None
             Random seed for drawing the samples with Gaussian errors.
-        batch_size : int, default=None
+        batch_size : int; default=None
             Size of batches in which to calculate posteriors. If None, all
             posteriors are calculated simultaneously. Simultaneous calculation
             is faster, but memory intensive for large data sets.
-        returnEnsemble : bool, default=False
+        returnEnsemble : bool; default=False
             If True, returns posterior for each flow in the ensemble as an
             array of shape (inputs.shape[0], N flows in ensemble, grid.size).
             If False, the posterior is averaged over the flows in the ensemble,
             and returned as an array of shape (inputs.shape[0], grid.size)
-        nan_to_zero : bool, default=True
+        nan_to_zero : bool; default=True
             Whether to convert NaN's to zero probability in the final pdfs.
 
         Returns
         -------
-        np.ndarray
+        jnp.ndarray
             For shape, see returnEnsemble description above.
         """
 
         # calculate posterior for each flow in the ensemble
-        ensemble = np.array(
+        ensemble = jnp.array(
             [
                 flow.posterior(
                     inputs=inputs,
@@ -311,20 +331,22 @@ class FlowEnsemble:
         )
 
         # re-arrange so that (axis 0, axis 1) = (inputs, flows in ensemble)
-        ensemble = np.rollaxis(ensemble, axis=1)
+        ensemble = jnp.rollaxis(ensemble, axis=1)
 
         if returnEnsemble:
             # return the ensemble of posteriors
             if normalize:
                 ensemble = ensemble.reshape(-1, grid.size)
-                ensemble = ensemble / np.trapz(y=ensemble, x=grid).reshape(-1, 1)
+                ensemble = ensemble / jnp.trapz(y=ensemble, x=grid).reshape(
+                    -1, 1
+                )
                 ensemble = ensemble.reshape(inputs.shape[0], -1, grid.size)
             return ensemble
         else:
             # return mean over ensemble
             pdfs = ensemble.mean(axis=1)
             if normalize:
-                pdfs = pdfs / np.trapz(y=pdfs, x=grid).reshape(-1, 1)
+                pdfs = pdfs / jnp.trapz(y=pdfs, x=grid).reshape(-1, 1)
             return pdfs
 
     def sample(
@@ -339,18 +361,18 @@ class FlowEnsemble:
 
         Parameters
         ----------
-        nsamples : int, default=1
+        nsamples : int; default=1
             The number of samples to be returned, either overall or per flow
             in the ensemble (see returnEnsemble below).
-        conditions : pd.DataFrame, optional
+        conditions : pd.DataFrame; optional
             If this is a conditional flow, you must pass conditions for
             each sample. nsamples will be drawn for each row in conditions.
-        save_conditions : bool, default=True
+        save_conditions : bool; default=True
             If true, conditions will be saved in the DataFrame of samples
             that is returned.
-        seed : int, optional
+        seed : int; optional
             Sets the random seed for the samples.
-        returnEnsemble : bool, default=False
+        returnEnsemble : bool; default=False
             If True, nsamples is drawn from each flow in the ensemble.
             If False, nsamples are drawn uniformly from the flows in the ensemble.
 
@@ -373,7 +395,7 @@ class FlowEnsemble:
             # if this isn't a conditional flow, sampling is straightforward
             if conditions is None:
                 # return nsamples drawn uniformly from the flows in the ensemble
-                N = int(np.ceil(nsamples / len(self._ensemble)))
+                N = int(jnp.ceil(nsamples / len(self._ensemble)))
                 samples = pd.concat(
                     [
                         flow.sample(N, conditions, save_conditions, seed)
@@ -390,7 +412,7 @@ class FlowEnsemble:
                     conditions = pd.concat([conditions] * nsamples)
 
                 # now the main sampling algorithm
-                seed = onp.random.randint(1e18) if seed is None else seed
+                seed = np.random.randint(1e18) if seed is None else seed
                 # if we are drawing more samples than the number of flows in
                 # the ensemble, then we will shuffle the conditions and randomly
                 # assign them to one of the constituent flows
@@ -400,27 +422,31 @@ class FlowEnsemble:
                         frac=1.0, random_state=int(seed / 1e9)
                     )
                     # split conditions into ~equal sized chunks
-                    chunks = onp.array_split(conditions_shuffled, len(self._ensemble))
+                    chunks = np.array_split(
+                        conditions_shuffled, len(self._ensemble)
+                    )
                     # shuffle the chunks
                     chunks = [
                         chunks[i]
                         for i in random.permutation(
-                            random.PRNGKey(seed), np.arange(len(chunks))
+                            random.PRNGKey(seed), jnp.arange(len(chunks))
                         )
                     ]
                     # sample from each flow, and return all the samples
                     return pd.concat(
                         [
-                            flow.sample(1, chunk, save_conditions, seed).set_index(
-                                chunk.index
+                            flow.sample(
+                                1, chunk, save_conditions, seed
+                            ).set_index(chunk.index)
+                            for flow, chunk in zip(
+                                self._ensemble.values(), chunks
                             )
-                            for flow, chunk in zip(self._ensemble.values(), chunks)
                         ]
                     ).sort_index()
                 # however, if there are more flows in the ensemble than samples
                 # being drawn, then we will randomly select flows for each condition
                 else:
-                    rng = onp.random.default_rng(seed)
+                    rng = np.random.default_rng(seed)
                     # randomly select a flow to sample from for each condition
                     flows = rng.choice(
                         list(self._ensemble.values()),
@@ -432,13 +458,18 @@ class FlowEnsemble:
                     return pd.concat(
                         [
                             flow.sample(
-                                1, conditions[i : i + 1], save_conditions, new_seed
+                                1,
+                                conditions[i : i + 1],
+                                save_conditions,
+                                new_seed,
+                            )
+                            for i, (flow, new_seed) in enumerate(
+                                zip(flows, seeds)
                             )
-                            for i, (flow, new_seed) in enumerate(zip(flows, seeds))
                         ],
                     ).set_index(conditions.index)
 
-    def save(self, file: str):
+    def save(self, file: str) -> None:
         """Saves the ensemble to a file.
 
         Pickles the ensemble and saves it to a file that can be passed as
@@ -458,10 +489,14 @@ class FlowEnsemble:
         save_dict = {
             "data_columns": self.data_columns,
             "conditional_columns": self.conditional_columns,
+            "latent_info": self.latent.info,
+            "data_error_model": self.data_error_model,
+            "condition_error_model": self.condition_error_model,
             "info": self.info,
             "class": self.__class__.__name__,
             "ensemble": {
-                name: flow._save_dict() for name, flow in self._ensemble.items()
+                name: flow._save_dict()
+                for name, flow in self._ensemble.items()
             },
         }
 
@@ -473,11 +508,13 @@ class FlowEnsemble:
         inputs: pd.DataFrame,
         epochs: int = 50,
         batch_size: int = 1024,
-        optimizer: Optimizer = None,
+        optimizer: Callable = None,
         loss_fn: Callable = None,
         convolve_errs: bool = False,
+        patience: int = None,
         seed: int = 0,
         verbose: bool = False,
+        progress_bar: bool = False,
     ) -> dict:
         """Trains the normalizing flows on the provided inputs.
 
@@ -486,26 +523,33 @@ class FlowEnsemble:
         inputs : pd.DataFrame
             Data on which to train the normalizing flows.
             Must have columns matching self.data_columns.
-        epochs : int, default=50
+        epochs : int; default=50
             Number of epochs to train.
-        batch_size : int, default=1024
+        batch_size : int; default=1024
             Batch size for training.
-        optimizer : jax Optimizer, default=adam(step_size=1e-3)
-            An optimizer from jax.experimental.optimizers.
-        loss_fn : Callable, optional
+        optimizer : optax optimizer
+            An optimizer from Optax. default = optax.adam(learning_rate=1e-3)
+            see https://optax.readthedocs.io/en/latest/index.html for more.
+        loss_fn : Callable; optional
             A function to calculate the loss: loss = loss_fn(params, x).
             If not provided, will be -mean(log_prob).
-        convolve_errs : bool, default=False
+        convolve_errs : bool; default=False
             Whether to draw new data from the error distributions during
-            each epoch of training. Assumes errors are Gaussian, and method
-            will look for error columns in `inputs`. Error columns must end
-            in `_err`. E.g. the error column for the variable `u` must be
-            `u_err`. Zero error assumed for any missing error columns.
-        seed : int, default=0
+            each epoch of training. Method will look for error columns in
+            `inputs`. Error columns must end in `_err`. E.g. the error column
+            for the variable `u` must be `u_err`. Zero error assumed for
+            any missing error columns. The error distribution is set during
+            ensemble instantiation.
+        patience : int; optional
+            Factor that controls early stopping. Training will stop if the
+            loss doesn't decrease for this number of epochs.
+        seed : int; default=0
             A random seed to control the batching and the (optional)
             error sampling.
-        verbose : bool, default=False
+        verbose : bool; default=False
             If true, print the training loss every 5% of epochs.
+        progress_bar : bool; default=False
+            If true, display a tqdm progress bar during training.
 
         Returns
         -------
@@ -514,22 +558,28 @@ class FlowEnsemble:
             in the ensemble.
         """
 
+        # generate random seeds for each flow
+        rng = np.random.default_rng(seed)
+        seeds = rng.integers(1e9, size=len(self._ensemble))
+
         loss_dict = dict()
 
-        for name, flow in self._ensemble.items():
+        for i, (name, flow) in enumerate(self._ensemble.items()):
 
             if verbose:
                 print(name)
 
             loss_dict[name] = flow.train(
-                inputs,
-                epochs,
-                batch_size,
-                optimizer,
-                loss_fn,
-                convolve_errs,
-                seed,
-                verbose,
+                inputs=inputs,
+                epochs=epochs,
+                batch_size=batch_size,
+                optimizer=optimizer,
+                loss_fn=loss_fn,
+                convolve_errs=convolve_errs,
+                patience=patience,
+                seed=seeds[i],
+                verbose=verbose,
+                progress_bar=progress_bar,
             )
 
         return loss_dict
